@@ -3,13 +3,13 @@
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
-# the MIT License: http://www.opensource.org/licenses/mit-license.php
+# the MIT License: https://www.opensource.org/licenses/mit-license.php
 """
 .. dialect:: postgresql+pypostgresql
     :name: py-postgresql
     :dbapi: pypostgresql
     :connectstring: postgresql+pypostgresql://user:password@host:port/dbname[?key=value&key=value...]
-    :url: http://python.projects.pgfoundry.org/
+    :url: https://python.projects.pgfoundry.org/
 
 .. note::
 
@@ -17,6 +17,13 @@
     integration** and may have unresolved issues.  The recommended PostgreSQL
     driver is psycopg2.
 
+.. deprecated:: 1.4 The py-postgresql DBAPI is deprecated and will be removed
+   in a future version. This DBAPI is superseded by the external
+    version available at external-dialect_. Please use the external version or
+    one of the supported DBAPIs to connect to PostgreSQL.
+
+.. TODO update link
+.. _external-dialect: https://github.com/PyGreSQL
 
 """  # noqa
 
@@ -45,6 +52,7 @@ class PGExecutionContext_pypostgresql(PGExecutionContext):
 class PGDialect_pypostgresql(PGDialect):
     driver = "pypostgresql"
 
+    supports_statement_cache = True
     supports_unicode_statements = True
     supports_unicode_binds = True
     description_encoding = None
@@ -69,6 +77,15 @@ class PGDialect_pypostgresql(PGDialect):
     def dbapi(cls):
         from postgresql.driver import dbapi20
 
+        # TODO update link
+        util.warn_deprecated(
+            "The py-postgresql DBAPI is deprecated and will be removed "
+            "in a future version. This DBAPI is superseded by the external"
+            "version available at https://github.com/PyGreSQL. Please "
+            "use one of the supported DBAPIs to connect to PostgreSQL.",
+            version="1.4",
+        )
+
         return dbapi20
 
     _DBAPI_ERROR_NAMES = [
